#==============================================================================
#
#  $Id$
#
"""
   Proactor based SSLServer implementation.

   The SSL data handler is designed to wrap another data handler, allowing you
   to easily SSLify any protocol with a data handler.
"""
#
#  Copyright (C) 2006 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#==============================================================================

from spug.io.proactor import DataHandler, ConnectHandler
from spug.util.log import debug, warn, error
from _minssl import *

class SSLDataHandler(DataHandler):

   ACCEPTING = 'ACCEPTING'
   CONNECTED = 'CONNECTED'
   SHUTDOWN_PENDING = 'SHUTDOWN_PENDING'
   CLOSED = 'CLOSED'

   def __init__(self, delegate, keyFile, certFile):
      """
	 parms:
	    delegate::
	       [@DataHandler] we pass all method invocations to this object as
	       soon as the SSL initialization sequence is established.
	    keyFile::
	       [string] name of the SSL private key file
	    certFile::
	       [string] name of the SSL certificate file
      """
      self.__delegate = delegate
      self.__output = ''

      # create two memory BIOs
      self.__readBIO = BIO_new(BIO_s_mem())
      self.__writeBIO = BIO_new(BIO_s_mem())

      # create the context and associate it with the certificate and private
      # key files
      self.__ctx = SSL_CTX_new(SSLv23_method())
      if SSL_CTX_use_certificate_file(self.__ctx, 'server.pem', 
                                      SSL_FILETYPE_PEM) != 1:
         raise Exception('Error loading certificate file')
      if SSL_CTX_use_PrivateKey_file(self.__ctx, 'server', 
                                     SSL_FILETYPE_PEM) != 1:
         raise Exception('Error loading private key')

      # create the SSL object, associated it with the read and write buffers
      self.__ssl = SSL_new(self.__ctx)
      SSL_set_bio(self.__ssl, self.__readBIO, self.__writeBIO)

      # put it into accept state
      SSL_set_accept_state(self.__ssl)
      self.__state = self.ACCEPTING
      self.__readyToRead = True

   # so here's the basic idea: 
   #
   # when we're connected, we want to just serve as an intermediary to the
   # delegate.  When we're in the ACCEPTING or any of the shutdown states, we
   # want to call the SSL_accept and SSL_shutdown functions respectively until
   # they tell us they're done.
   #
   # The sole channel for communication with the rest of the functions is the
   # read and write BIOs.  __output is used as a staging area for data that
   # has been peeked but not gotten, but that is above the level of the state
   # machine.

   def __process(self):
      debug('state = %s' % self.__state)
      if self.__state == self.ACCEPTING:
         self.__readyToRead = False
         rc = SSL_accept(self.__ssl)
         if rc == 1:
            self.__state = self.CONNECTED
         else:
            rc = SSL_get_error(self.__ssl, rc)
            if rc == SSL_ERROR_WANT_READ:
               self.__readyToRead = True
               return
            elif rc == SSL_ERROR_SSL:
               # protocol error - log an error message and drop the connection
               error('Protocol error')
               self.__output = ''
               self.__state = self.CLOSED
            else:
               raise Exception('got ssl error: %d' % rc)

      if self.__state == self.CONNECTED:

         # if we're connected, everything falls through to the delegate

         if self.__delegate.readyToClose():
            self.__state = self.SHUTDOWN_PENDING

         else:
            # instead of 1024, should use the last MTU size received from the
            # proactor

            # feed the delegate data from the SSL object until either we run
            # out or he can't take no more
            data = True
            while self.__delegate.readyToPut():
               data = SSL_readx(self.__ssl, 1024)
               if data == SSL_ERROR_WANT_READ:
                  self.__readyToRead = True
                  break
               else:
                  assert data
                  debug('putting data %s' % repr(data))
                  self.__delegate.put(data)

            if self.__delegate.readyToGet() and \
               BIO_ctrl_pending(self.__writeBIO) < 1024:

               # pull the datafrom the delegate
               data = self.__delegate.peek(1024)
               self.__delegate.get(len(data))
               debug('got data %s' % repr(data))
               
               # push it through SSL
               SSL_write(self.__ssl, data)

      if self.__state == self.SHUTDOWN_PENDING:

         self.__readyToRead = False
         rc = SSL_shutdown(self.__ssl)
         if rc > -1:
            # if we've pushed everything to the client, we're done
            if BIO_ctrl_pending(self.__writeBIO) == 0:
               self.__state = self.CLOSED
         else:
            raise Exception('got error during shutdown %d' % rc)

   def readyToGet(self):
      return BIO_ctrl_pending(self.__writeBIO) and True or False

   def readyToPut(self):
      return self.__readyToRead

   def readyToClose(self):
      return not self.__output and self.__state == self.CLOSED

   def handleConnect(self):
      pass

   def peek(self, size):
      self.__process()
      if len(self.__output) < size:
         self.__output += BIO_readx(self.__writeBIO, size - len(self.__output))
      data = self.__output[:size]
      debug('peeking %s' % repr(data))
      return data

   def get(self, size):
      self.__process()
      self.__output = self.__output[size:]

   def put(self, data):
      debug('putting %s' % repr(data))
      BIO_write(self.__readBIO, data)
      self.__process()

   def __del__(self):
      # clean up all of the allocated SSL objects (we don't need to free the
      # BIO's - this will be done by the SSL object)
      SSL_free(self.__ssl)
      SSL_CTX_free(self.__ctx)

   if 0:
      def readyToPut(self):
         if self.__state == self.ACCEPTING:
            return self.__readyToRead
         else:
            return self.__delegate.readyToPut()

      def readyToGet(self):
         # we're ready to get if there's data in the write buffer or we're
         # connected and the delegate is ready to get
         if BIO_ctrl_pending(self.__writeBIO) > 0 or \
            self.__state == self.CONNECTED and self.__delegate.readyToGet():
            return True
         else:
            return False

      def readyToClose(self):
         if self.__state == self.SHUTDOWN_SENT:
            return True
         elif self.__delegate.readyToClose():
            SSL_shutdown(self.__ssl)

            # transfer any data to the output buffer
            data = True
            while data:
               data = BIO_readx(self.__writeBIO, 1024)
               self.__output += data

            self.__state = self.SHUTDOWN_PENDING1

      def handleConnect(self):
         pass

      def peek(self, size):
         # fill the output staging area (which will get wiped out in the
         # subsequent get)
         if not self.__output:
            # if there's no data in the buffer, pull from the delegate.  The only
            # way we should have returned true from readyToGet is if we are in
            # accept state and there was data in the buffer or if we are
            # connected and the delegate was ready to get.
            if not BIO_ctrl_pending(self.__writeBIO) and \
               self.__state == self.CONNECTED:

               data = self.__delegate.peek(size)
               self.__delegate.get(len(data))
               SSL_write(self.__ssl, data)

            self.__output = BIO_readx(self.__writeBIO, size)

            if not self.__output and self.__state == self.SHUTDOWN_PENDING1:
               self.__state = self.SHUTDOWN_PENDING2

         debug('writing %s' % repr(self.__output))
         return self.__output

      def get(self, size):
         self.__output = self.__output[size:]
         if not self.__output and self.__state == self.SHUTDOWN_PENDING2:
            self.__state = self.SHUTDOWN_SENT

      def put(self, data):
         debug('reading %s' % repr(data))
         BIO_write(self.__readBIO, data)

         # if we're accepting, call accept
         if self.__state == self.ACCEPTING:
            self.__readyToRead = False
            rc = SSL_accept(self.__ssl)
            if rc == 1:
               self.__state = self.CONNECTED
            else:
               rc = SSL_get_error(self.__ssl, rc)
               if rc != SSL_ERROR_WANT_READ:
                  raise Exception('got ssl error: %d' % rc)
               else:
                  self.__readyToRead = True

         # if we're connecting, read from the SSL object and pass it back to the
         # delegate
         else:
            data = True
            while data:
               data = SSL_readx(self.__ssl, 1024)
               if isinstance(data, str):
                  self.__delegate.put(data)
                  if not self.__delegate.readyToPut():
                     data = False
               else:
                  debug('got non string data: %s' % data)
                  break


      def handleRead(self, dispatcher):
         data = self._sock.read(MTU)

class SSLConnectHandler(ConnectHandler):
   """
      Simple SSL HTTP Server implementation.
   """

   def __init__(self, delegateFactory, keyFile, certFile):
      """
         parms:
            delegateFactory::
               [callable<>] function that returns a delegate DataHandler
               for a new connection.  This will actually be the final producer
               and consumer of data passed through an @SSLDataHandler.
	    keyFile::
	       [string] name of the SSL private key file
	    certFile::
	       [string] name of the SSL certificate file
      """

      self.__delegateFactory = delegateFactory
      self.__shutdown = False
      self.__keyFile = keyFile
      self.__certFile = certFile

   def handleConnect(self, connection):
      connection.dataHandler = \
	 SSLDataHandler(self.__delegateFactory(), keyFile, certFile)

   def shutdown(self):
      self.__shutdown = True

   def readyToClose(self):
      return self.__shutdown

